@@ -1,4 +1,4 @@
-// Copyright 2021 the Deno authors. All rights reserved. MIT license.
+// Copyright 2018-2022 the Deno authors. All rights reserved. MIT license.
 
 use deno_core::error::bad_resource_id;
 use deno_core::error::generic_error;
@@ -57,6 +57,7 @@ struct Symbol {
   result_type: NativeType,
 }
 
+#[allow(clippy::non_send_fields_in_send_ty)]
 unsafe impl Send for Symbol {}
 unsafe impl Sync for Symbol {}
 
@@ -78,18 +79,21 @@ impl Resource for DynamicLibraryResource {
 impl DynamicLibraryResource {
   fn register(
     &mut self,
-    symbol: String,
+    name: String,
     foreign_fn: ForeignFunction,
   ) -> Result<(), AnyError> {
+    let symbol = match &foreign_fn.name {
+      Some(symbol) => symbol,
+      None => &name,
+    };
     // By default, Err returned by this function does not tell
     // which symbol wasn't exported. So we'll modify the error
     // message to include the name of symbol.
-    let fn_ptr = match unsafe { self.lib.symbol::<*const c_void>(&symbol) } {
+    let fn_ptr = match unsafe { self.lib.symbol::<*const c_void>(symbol) } {
       Ok(value) => Ok(value),
       Err(err) => Err(generic_error(format!(
         "Failed to register symbol {}: {}",
-        symbol,
-        err.to_string()
+        symbol, err
       ))),
     }?;
     let ptr = libffi::middle::CodePtr::from_ptr(fn_ptr as _);
@@ -103,7 +107,7 @@ impl DynamicLibraryResource {
     );
 
     self.symbols.insert(
-      symbol,
+      name,
       Symbol {
         cif,
         ptr,
@@ -126,6 +130,11 @@ pub fn init<P: FfiPermissions + 'static>(unstable: bool) -> Extension {
       ("op_ffi_load", op_sync(op_ffi_load::<P>)),
       ("op_ffi_call", op_sync(op_ffi_call)),
       ("op_ffi_call_nonblocking", op_async(op_ffi_call_nonblocking)),
+      ("op_ffi_call_ptr", op_sync(op_ffi_call_ptr)),
+      (
+        "op_ffi_call_ptr_nonblocking",
+        op_async(op_ffi_call_ptr_nonblocking),
+      ),
       ("op_ffi_ptr_of", op_sync(op_ffi_ptr_of::<P>)),
       ("op_ffi_buf_copy_into", op_sync(op_ffi_buf_copy_into::<P>)),
       ("op_ffi_cstr_read", op_sync(op_ffi_cstr_read::<P>)),
@@ -319,7 +328,7 @@ fn value_as_f64(value: Value) -> Result<f64, AnyError> {
   }
 }
 
-#[derive(Serialize, Deserialize, Debug)]
+#[derive(Serialize, Deserialize, Debug, Clone, Copy)]
 struct U32x2(u32, u32);
 
 impl From<u64> for U32x2 {
@@ -337,6 +346,7 @@ impl From<U32x2> for u64 {
 #[derive(Deserialize, Debug)]
 #[serde(rename_all = "camelCase")]
 struct ForeignFunction {
+  name: Option<String>,
   parameters: Vec<NativeType>,
   result: NativeType,
 }
@@ -464,6 +474,49 @@ struct FfiCallArgs {
   buffers: Vec<Option<ZeroCopyBuf>>,
 }
 
+#[derive(Deserialize)]
+#[serde(rename_all = "camelCase")]
+struct FfiCallPtrArgs {
+  pointer: U32x2,
+  def: ForeignFunction,
+  parameters: Vec<Value>,
+  buffers: Vec<Option<ZeroCopyBuf>>,
+}
+
+impl From<FfiCallPtrArgs> for FfiCallArgs {
+  fn from(args: FfiCallPtrArgs) -> Self {
+    FfiCallArgs {
+      rid: 0,
+      symbol: String::new(),
+      parameters: args.parameters,
+      buffers: args.buffers,
+    }
+  }
+}
+
+impl FfiCallPtrArgs {
+  fn get_symbol(&self) -> Symbol {
+    let fn_ptr: u64 = self.pointer.into();
+    let ptr = libffi::middle::CodePtr::from_ptr(fn_ptr as _);
+    let cif = libffi::middle::Cif::new(
+      self
+        .def
+        .parameters
+        .clone()
+        .into_iter()
+        .map(libffi::middle::Type::from),
+      self.def.result.into(),
+    );
+
+    Symbol {
+      cif,
+      ptr,
+      parameter_types: self.def.parameters.clone(),
+      result_type: self.def.result,
+    }
+  }
+}
+
 fn ffi_call(args: FfiCallArgs, symbol: &Symbol) -> Result<Value, AnyError> {
   let buffers: Vec<Option<&[u8]>> = args
     .buffers
@@ -558,6 +611,26 @@ fn ffi_call(args: FfiCallArgs, symbol: &Symbol) -> Result<Value, AnyError> {
   })
 }
 
+fn op_ffi_call_ptr(
+  _state: &mut deno_core::OpState,
+  args: FfiCallPtrArgs,
+  _: (),
+) -> Result<Value, AnyError> {
+  let symbol = args.get_symbol();
+  ffi_call(args.into(), &symbol)
+}
+
+async fn op_ffi_call_ptr_nonblocking(
+  _state: Rc<RefCell<deno_core::OpState>>,
+  args: FfiCallPtrArgs,
+  _: (),
+) -> Result<Value, AnyError> {
+  let symbol = args.get_symbol();
+  tokio::task::spawn_blocking(move || ffi_call(args.into(), &symbol))
+    .await
+    .unwrap()
+}
+
 fn op_ffi_call(
   state: &mut deno_core::OpState,
   args: FfiCallArgs,
